// SerialServer.cpp : Defines the entry point for the application.
// Copyright (C) 2003 Christian Ghisler, C. Ghisler & Co.
// This program is copyrighted freeware
// Kommandozeilenversion als Ausgangspunkt fr grere Arduinos,
// heha, 1902xx

#include <windows.h>
#include <stdio.h>
#include "blockio.h"

// Global Variables:
char renamefromname[MAX_PATH];

// ****************** Main *********************
// *** Main loop to receive and handle commands from client ***
// ************************************************************

int mainCRTStartup() {
 int portno=0;
 int portspeed=115200;
 int argc;
 PWSTR*argv=CommandLineToArgvW(GetCommandLineW(),&argc);
 if (argc>1) portno=_wtoi(argv[1]);
 if (argc>2) portspeed=_wtoi(argv[2]);
 if (!portspeed) printf("Wrong baudrate");
 if (argc>3) printf("superfluous argument(s) ignored\n");
 Err err;
 uint8_t ch;
 HANDLE FileHandle;
 char FileName[260];
 char* p;
 uint32_t restartat;
 char* block;
 char cmd;
 bool isadir,ok,overwritefile;

 SetErrorMode(0x08001);  // don't show errors when access to drive fails!
 restartat=0;
 overwritefile=false;
 block=NULL;

 if (!SerialConnect(portno,portspeed)) return 5;
		// Try to allocate big buffer - if not possible, use internal(slower)
 block=new char[64*1024];
 if (!block) {
  printf("Not enough memory to allocate buffer\n");
  return 4;
 }
		// Main command loop starts here!
 for(;;) {
  ch=RQ_TRANSFERERR;
  printf("Ready...");
			// Wait for next command from server
  FileName[0]=0;
  cmd=ReceiveNextBlock(FileName,sizeof FileName-1,NULL,NULL);
  switch (cmd) {
   case CMD_CRCERR: SerialSendChar(RQ_CRCERR); break;
   case CMD_ABORT: goto breakloop;				// user-abort
   case CMD_CONNECT:			// sort of ping to check that the server is responding
    printf("Connect!");
    SerialSendChar(27/*RQ_OK*/);
    break;
   case 'L':
    isadir=false;
    if (FileName[0]=='\\' && FileName[1]==0) {		// the root
     if (restartat) {
      printf("Resume not supported!");	// no resume with dirs!
      ch=RQ_TRANSFERERR;
     }else{
      printf("GET valid...");
      ch=RQ_OK;
     }
     FileHandle=0;
    }else{
     if (strlen(FileName)>4) {  // drives are OK
      HANDLE fh;
      WIN32_FIND_DATA srec;
      char*e=FileName+strlen(FileName)-1;
      if (*e=='\\') *e=0;
      fh=FindFirstFile(FileName+1,&srec);
      if (fh!=INVALID_HANDLE_VALUE) {
       isadir=!!(srec.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);
       FindClose(fh);
      }
     }else isadir=true;
     if (!isadir) {
      printf("LIST invalid...");
      ch=RQ_TRANSFERERR;
     }else{
      printf("LIST valid...");
      ch=RQ_OK;
     }
    }
    SerialSendChar(ch);
    if (ch==RQ_OK) {
     printf("name:%s\n",FileName);
     err=SerialSendFile(INVALID_HANDLE_VALUE,FileName,block,restartat,FALSE);
    }else if (err) {
     printf("Error %d in list",err);
    }else{
     printf(" (not found)");
    }
    break;
   case CMD_GETSIZE:		// client wants to get only file size
   case CMD_GET:		// client tries to get file or dir
    isadir=false;
    if (FileName[0]=='\\' && FileName[1]==0) {		// the root
     printf("\\ is not a file!...");
     ch=RQ_TRANSFERERR;
     FileHandle=INVALID_HANDLE_VALUE;
    }else{
     FileHandle=CreateFile(FileName+1,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
     if (FileHandle!=INVALID_HANDLE_VALUE && restartat>0) {
      if (SetFilePointer(FileHandle,restartat,NULL,FILE_BEGIN)==-1) {
       CloseHandle(FileHandle);
       FileHandle=INVALID_HANDLE_VALUE;
      }
     }
     if (FileHandle==INVALID_HANDLE_VALUE) {
      printf("GET invalid...");
      ch=RQ_TRANSFERERR;
     }else{
      ch=RQ_OK;
     }
    }
    SerialSendChar(ch);
    if (ch==RQ_OK) {
     printf("name:%s\n",FileName);
     err=SerialSendFile(FileHandle,FileName,block,restartat,cmd==CMD_GETSIZE);
    }else if (err) printf("Error %d in send",err);
    if (FileHandle!=INVALID_HANDLE_VALUE) CloseHandle(FileHandle);
    FileHandle=INVALID_HANDLE_VALUE;
    break;
   case CMD_SEND:		// client sends file or dir
    if (restartat||overwritefile) FileHandle=INVALID_HANDLE_VALUE;
    else FileHandle=CreateFile(FileName+1,
      GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
    if (FileHandle!=INVALID_HANDLE_VALUE) {
     CloseHandle(FileHandle);
     FileHandle=INVALID_HANDLE_VALUE;
     ch=RQ_FAILED;
    }else{
     FileHandle=CreateFile(FileName+1,
       GENERIC_WRITE,0,0,
       restartat?OPEN_EXISTING:CREATE_ALWAYS,
       FILE_ATTRIBUTE_NORMAL,0);
     err = FileHandle==INVALID_HANDLE_VALUE || restartat && SetFilePointer(FileHandle,restartat,NULL,FILE_BEGIN)==-1;
     if (err) {
      printf("PUT invalid...");
      ch=RQ_TRANSFERERR;
     }else ch=RQ_OK;
    }
    SerialSendChar(ch);
    if (ch==RQ_OK) {
     printf("name:%s\n",FileName);
     err=SerialReceiveFile(FileHandle, FileName, block, restartat);
    }else if (err) printf("Error %d in recv",err);
    if (FileHandle!=INVALID_HANDLE_VALUE) CloseHandle(FileHandle);
    FileHandle=INVALID_HANDLE_VALUE;
    break;
   case CMD_DELETE:		// delete file or dir
    printf("name:%s\n",FileName);
    ok=!!DeleteFile(FileName+1);
    if (!ok) ok=!!RemoveDirectory(FileName+1);
    if (ok) ch=RQ_OK;
    SerialSendChar(ch);
    break;
   case CMD_MKDIR:		// delete file or dir
    if (FileName[0]=='\\' && FileName[2]==':') {
     printf("name:%s\n",FileName);
     if (CreateDirectory(FileName+1,NULL)) ch=RQ_OK;
    }
    SerialSendChar(ch);
    break;
   case CMD_RENFR:		// renamefrom file or dir
    if (FileName[0]=='\\' && FileName[2]==':' && FileName[3]=='\\') {
     strncpy(renamefromname,FileName,sizeof renamefromname-1);
     ch=RQ_OK;
    }
    SerialSendChar(ch);
    break;
   case CMD_RENTO:		// rename to file or dir
   case CMD_COPYTO:		// copy to file or dir
    if (renamefromname[0]=='\\' && FileName[0]=='\\' && FileName[2]==':' && FileName[3]=='\\') {
     SerialSendChar(RQ_WAIT);	//Server sends RQ_WAIT for long operations
     if (cmd==CMD_COPYTO) {
      if (CopyFile(renamefromname+1,FileName+1,true)) ch=RQ_OK; // only same dir!
      else ch=RQ_FAILED;
     }else{
      if (MoveFile(renamefromname+1,FileName+1)) ch=RQ_OK; // only same dir!
      else ch=RQ_FAILED;
     }
    }
    SerialSendChar(ch);
    break;
   case CMD_SETATTR:
    p=strchr(FileName,'\t');
    if (p && FileName[0]=='\\' && FileName[2]==':' && FileName[3]=='\\') {
     *p++=0;
     uint32_t attributes=atoi(p);
     printf("name:%s\n",FileName);
     if (SetFileAttributes(FileName+1,attributes)) ch=RQ_OK;
    }
    SerialSendChar(ch);
    break;
   case CMD_SETTIME:
    p=strchr(FileName,'\t');
    if (p && FileName[0]=='\\' && FileName[2]==':' && FileName[3]=='\\') {
     *p++=0;
     FILETIME FileDate;
     if (ConvertIsoDateToDateTime(p,FileDate)) {
      printf("name:%s\n",FileName);
      FileHandle=CreateFile(FileName+1,GENERIC_WRITE,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
      if (FileHandle!=INVALID_HANDLE_VALUE) {
       if (SetFileTime(FileHandle,NULL,NULL,&FileDate)) ch=RQ_OK;
       CloseHandle(FileHandle);
      }
     }
    }
    SerialSendChar(ch);
    break;
   case CMD_RESTART:
    restartat=atoi(FileName);
    SerialSendChar(RQ_OK);
    break;
   case CMD_OVERWRITE:
    overwritefile=true;
    SerialSendChar(RQ_OK);
    break;
   default:
    SerialSendChar(RQ_FAILED);
  }
  if (cmd!=CMD_RESTART && cmd!=CMD_CRCERR) restartat=0;	// only use restart for next command!
  if (cmd!=CMD_OVERWRITE && cmd!=CMD_CRCERR) overwritefile=false;	// only use overwrite for next command!
 }
breakloop:
 if (block) delete[] block;
 SerialClose();
 return 0;
}
